#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import colormap
import archive
import util

# import the four horsemen of the bitpocalypse:
from util import bits_pack, ncbits_pack, ncbits_of, bits_of, bitstruct_pack

def DelvImageFactory(src, *args, **kwargs):
    """Return the appropriate kind of DelvImage subclass for the 
       delv.archive.Resource object you provide."""
    return _CLASS_HINTS.get(src.subindex, DelvImage)(src, *args, **kwargs)
    
def DelvImageFactoryMode(src, mode, *args, **kwargs):
    """Return the appropriate kind of DelvImage subclass for the 
       specified mode, one of tiles, portrait, landscape or sized."""
    return _NAME_HINTS.get(mode, DelvImage)(src, *args, **kwargs)

class UnknownOpcode(Exception): pass

class DelvImage(object):
    """This is the base class for all forms of Delver Compressed (sprite)
       graphics, images in the proprietary format used by the Delver Engine.
       You should probably not instantiate this class directly, but instead
       one of its child classes (General, for sized images; TileSheet, for
       32x512 sets of sixteen 32x32 tiles; Portrait, for 64x64 character
       portraits; or Landscape, for 288x32 level images. You can use 
       DelvImageFactory (arguments same as DelvImage's constructor) to 
       automatically pick the right class, if you are loading from a
       delv.archive.Resource object.
       
       The file format used, is documented here in detail:
       http://www.ferazelhosting.net/wiki/Delver%20Compressed%20Graphics
    """
    canonical_size = 256,256
    has_header = False
    def __init__(self,src=None,width=256,height=256,flags=0,logical_width=256):
        """Create a new Delver Compressed Graphics Image. src can be None,
           the default, which creates an empty image of the default size
           for this type, or an random-access indexable item such as a list,
           string or array, which will be assumed to be compressed data, 
           or a delv Resource object (also assumed to be compressed data.)"""
        if isinstance(src, archive.Resource):
            self.src = bytearray(src.get_data())
        elif hasattr(src, "__getitem__"):
            self.src = bytearray(src)
        else:
            self.src = None
        
        
        if self.has_header and self.src:
            header = self.src[0:4]
            self.src = self.src[4:]
            data_cursor = 0
            self.width = self.logical_width = bits_of(header, 8+6, 0) << 2
            self.flags = bits_of(header, 2,  14)
            self.height = bits_of(header, 15, 16) << 1
            self.flags2 = bits_of(header, 1, 31)
            self.logical_height = self.height + self.flags2
            # TODO - figure out what the deal is with this.
            # DelvTechWiki conjectures that it has something to do with 
            # objects that have some sort of response to dragging/clicking
            # only in certain areas. But it has colors similar to the main
            # object...
            if self.flags: 
                self.logical_width += 4
                self.width += self.flags
            if self.flags2:
                self.height += self.flags2
                # interpreting 8F0C correctly seems to be mutually exclusive
                # with 8F1A. They have the same identified or suspected flag
                # bits. One possibility is that 8F0C is an earlier version
                # of the graphics format (I at least do not recognize the pot
                # as appearing in the current game); another is that the size
                # information is encoded somewhere else e.g. in a script.
                # Neither of the two resources mentioned above contain any
                # unknown opcodes (and that would be a really obtuse way
                # to encode size information anyhow.)
        elif self.has_header and not self.src:
            self.width,self.height = width,height
            self.flags = 0
            data_cursor = 0
            self.logical_width = logical_width
            self.logical_height = self.height
        else:
            self.width,self.height = self.canonical_size
            self.flags = 0
            data_cursor = 0
            self.logical_width = self.width
            self.logical_height = self.height
        self.cursor = 0
        self.image = bytearray(self.logical_width * self.logical_height)
        if src: self.decompress(self.src, data_cursor)
        self.cached_visual = None
    def decompress(self, data, cursor):
        """Decompress the indexable-item data provided into this image.
           You shouldn't normally need to call this explicitly."""
        while cursor < len(data):
            opcode = data[cursor]
            if opcode < 0x80:
                # short copy operation 0x00-0x7F
                operation = data[cursor:cursor+2]; cursor += 2
                index =-(ncbits_of(operation, (3,8),  (7,1)) + 1)
                length =   bits_of(operation,  3, 13) + 3
                literals = bits_of(operation,  2, 11)
                self.data(data[cursor:cursor+literals]); cursor += literals
                self.copy(length,index)
            elif opcode < 0xC0:
                # long copy operation 0x80-0xBF
                operation = data[cursor:cursor+3]; cursor += 3
                index =-(ncbits_of(operation, (6,16), (3,8), (6,2)) + 1)
                length =   bits_of(operation,  5,11) + 3
                literals = bits_of(operation,  2,22)
                self.data(data[cursor:cursor+literals]); cursor += literals
                self.copy(length,index)
            elif opcode < 0xD0:
                # pixel data 0xC0-0xCF
                operation = data[cursor:cursor+1]; cursor += 1
                size =    (bits_of(operation,  4,4) + 1) * 4
                self.data(data[cursor:cursor+size]); cursor += size
            elif opcode < 0xE0:
                # Short data
                # They do not seem to have any visual effect.
                # It is quite possible that this represents a short run of 
                # literal data (< C0); it only appears as the penultimate
                # opcode in the corpus. On this basis we assume 4 bits of
                # literals, but only D2 is seen.
                operation = data[cursor:cursor+1]; cursor += 1
                literals =  bits_of(operation,  4,4)
                self.data(data[cursor:cursor+literals]); cursor += literals
                #print "Unknwn opcode %02X lit=%d c=%X"%(operation[0],literals,
                #    cursor)
            elif opcode < 0xF0:
                # short run 0xE0-0xEF
                operation = data[cursor:cursor+2]; cursor += 2
                length =   bits_of(operation,  4,4) + 3
                color =    operation[1]
                self.run(length,color)
            elif opcode == 0xF0:
                # long run 0xF0
                operation = data[cursor:cursor+3]; cursor += 3
                length =   operation[1] + 3
                color =    operation[2]
                self.run(length,color)
            elif opcode == 0xFF:
                # terminate 0xFF
                operation = data[cursor]; cursor += 1
                #print "Orderly termination. dc:%x pc:%X"%(cursor,self.cursor)
            else:
                # unknown opcode
                raise UnknownOpcode("0x%02X at 0x%06X"%(opcode,cursor))

    def compress(self, image):
        """Create the compressed version of the graphic. You shouldn't
           normally need to call it explicitly."""
        # The basic idea here is that we have each opcode make a "bid" on
        # the image data at the current cursor position, then we use a greedy
        # approach to pick the bidder who gives the best (local) compression
        # ratio.
        # Each cursor position needs to be bid on by the following operations:
        # Short Data (Dx)               Long Data (Cx)
        # Short Copy with 0 literals    Long Copy with 0 literals
        # Short Copy with 1 literal     Long Copy with 1 literal
        # Short Copy with 2 literals    Long Copy with 2 literals
        # Short Run (Ex)                Long Run (F0)
        operations = [
                      lambda i,d: self.en_short_data(i,d),
                      lambda i,d: self.en_long_data(i,d),
                      lambda i,d: self.en_short_copy(i,d,0),
                      lambda i,d: self.en_short_copy(i,d,1),
                      lambda i,d: self.en_short_copy(i,d,2),
                      lambda i,d: self.en_long_copy(i,d,0),
                      lambda i,d: self.en_long_copy(i,d,1),
                      lambda i,d: self.en_long_copy(i,d,2),
                      lambda i,d: self.en_short_run(i,d),
                      lambda i,d: self.en_long_run(i,d)
        ]
        codes = []
        icursor = 0
        while icursor < len(self.image):
            # each of the en_* codes returns a compression_ratio, end_cursor,
            # code tuple. We pick the best one until we're done.
            _,icursor,code = max([op(icursor,image) for op in operations])
            codes.append(code)
            #if not code[0]&0x80: break
        # The scheme above will tend to produce lots of sequences like
        # C0 aa bb cc dd, C0 ee ff gg hh, C0 ii jj kk ll
        # and we need to condense them into
        # C2 aa bb cc dd ee ff gg hh ii jj kk ll
        # which is what condense does. The whole reason for this nonsense is
        # that we don't want greedy behavior in the literal data opcodes.

        if self.has_header:
            data = bytearray(4)
            bits_pack(data, self.width>>2, 14, 0 )
            bits_pack(data, self.flags,     2, 15) 
            bits_pack(data, self.height,   16, 16)
        else: 
            data = bytearray()
        data += self.condense_opcodes(codes)
        data += '\xFF' # Termination opcode
        return data

    def get_data(self):
        """Return the compressed image data, as a bytearray."""
        if not self.src: self.src = self.compress(self.image)
        return self.src

    def get_logical_image(self):
        """Get the whole image. Returns a one-dimensional
           bytearray. This will include data not part of the image
           displayed to the user, if it exists (the purpose of that data
           is unknown at this time.)"""
        return self.image
    def get_from(self, x,y,w,h):
        """Get a section of the image, bounded by the rectangle (x,y,w,h).
           It is returned as a flat bytearray, size w*h."""
        gf = bytearray()
        for yr in xrange(y,y+h):
            gf += self.image[yr*self.logical_width+x:yr*self.logical_width+x+w]
        return gf
    def get_image(self):
        """Get only the part of the image normally displayed by 
           the engine to the user. This is usually all of the image;
           except for certain Sized image resources ID 8Fxx."""
        if self.width == self.logical_width: 
            self.cached_visual = self.image
        if self.cached_visual: return self.cached_visual
        self.cached_visual = bytearray(self.width*self.height)
        v_cursor = 0
        for x in xrange(self.height):
            self.cached_visual[v_cursor:v_cursor+self.width] = (
                self.image[x*self.logical_width:x*self.logical_width+self.width
                ])
            v_cursor += self.width
        return self.cached_visual
    def set_image(self, data):
        """Set the uncompressed data of the logical image. data must be a 
           linear indexable item of size logical_width*logical_height, of
           integer values that represent indexed color pixels. Left to right,
           top to bottom, densely packed."""
        self.image = bytearray(data)
        self.cached_visual = None
        self.src = None
    def draw_into(self,src,x=0,y=0,w=0,h=0):
        """Draw an image src into this object, optionally specifying
           destination coordinates and a width and height if you wish
           to copy less than the entirety of the source image.
           src must be a flat indexable of size w*h.
           The source must already be 8-bit indexed color using the colormap
           for your target game (e.g the Cythera CLUT); dithering
           and color-matching are beyond the scope of this project."""
        self.src = None
        if not self.image: 
             self.image = bytearray(self.logical_width*self.logical_height)
        for yr in xrange(y,y+h): 
             self.image[yr*self.logical_width+x:yr*self.logical_width+x+w] = (
                 src[(yr-y)*w:(yr-y)*w + w])
    def draw_into_tile(self,src,n):
        """Conveninence method to draw to a particular tile. n has the
           same semantics as for get_tile below, including for situations
           in which non-canonical tileset shapes are addressed."""
        self.draw_into(*(src,)+self.tile_rect(n))
    def get_subtile(self,n,m):
        """Get the mth subtile of the nth tile of this image. (See .get_tile
           for definition of nth tile). The mth subtile is the 8x8 chunk of
           the tile, numbered starting from the top-left, but going 
           top-to-bottom THEN left-to-right."""
        return self.get_from(*self.subtile_rect(n,m))
    def get_tile(self,n):
        """Get the nth 32x32 tile of this image. Clasically meaningful only
           for tile sheets, where it returns the nth from the top (counting
           from zero.) For the sake of completeness, though, it will work
           on other images, in which case the tiles are numbered top to 
           bottom, left to right. The format is the same as get_from."""
        return self.get_from(*self.tile_rect(n))

    def tile_rect(self,n):
        """Return the bounding rectangle of the nth tile."""
        y = (n*32)%512
        x = (n//16)*32
        return (x,y,32,32)

    def subtile_rect(self,n,m):
        return ((n//16)*32 + (m//4)*8,
                (n*32)%512 + (m%4)*8,
                8,8)
  
    def get_size(self):
        """Return (width,height)."""
        return self.width,self.height
    ##########################  PRIVATE METHODS  #######################
    # The methods below may change freely between delv versions, and are
    # intended for the internal use of other delv code only.
    def run(self, length, color):
        maxcursor = self.cursor + length
        while self.cursor < maxcursor:
            self.image[self.cursor] = color
            self.cursor += 1
    def copy(self, length, origin):
        abs_origin = self.cursor + origin
        copy_width = -origin
        for n in xrange(length):
            self.image[self.cursor] = self.image[abs_origin + n%copy_width]
            self.cursor += 1
    def data(self, pixels):
        self.image[self.cursor:self.cursor+len(pixels)] = pixels
        self.cursor += len(pixels)
    def put_pixel(self, color):
        self.image[self.cursor] = color
        self.cursor += 1
    # encoder methods
    # All of them return compression_ratio, end_cursor_position, encoded_data
    # tuples.
    # compression_ratio = lenth of image data / length of encoded data
    def en_short_data(self,i,d):
        end = min(i+3,len(d))
        datalen = end-i
        code = bytearray(1+datalen)
        bits_pack(code, 0xD,          4, 0)
        bits_pack(code, datalen,      2, 6)
        for n,m in zip(xrange(i,i+datalen),xrange(1,datalen+1)):
            code[m]=d[n]
        return datalen/(datalen+1.0),end,code
    def en_long_data(self,i,d):
        if len(d)-i < 4:
            return -1,i,''
        return 4/5.0,i+4,'\xC0'+d[i:i+4]
    def searchback(self, d, i, lits, start, stop, mlen,llen=3):
        found = -1
        match_found = -1
        lowest_clen = llen
        highest_clen = min(mlen,len(d)-i)
        found_clen = 0
        while (highest_clen >= lowest_clen):
            mid_clen = (highest_clen-lowest_clen)/2 + lowest_clen
            found = d.rfind(d[i+lits:i+lits+mid_clen],start,stop)
            if found >= 0:
                lowest_clen = mid_clen+1
                found_clen = mid_clen
                match_found = found
            else:
                highest_clen = mid_clen-1
        return match_found,found_clen if match_found >= 0 else llen
    def en_short_copy(self,i,d,lits):
        #if lits: return -1,i,''
        found,clen = self.searchback(d,i,lits,max(0,i-(1024-lits)),i+lits,10)
        if found < 0:
            return -1,i,''

        index = -(found-i+1)+lits

        code = bytearray(2)
        bitstruct_pack(code, [
            (0,      [(1,0)       ]),
            (index,  [(3,8), (7,1)]),
            (clen-3, [(3,13)      ]),
            (lits,   [(2,11)      ])])
          
        #check = ncbits_of(code, (3,8), (7,1))
        #if check != index:
        #    print "ERROR %08X %08X"%(index,check), clen-3, lits
        code += d[i:i+lits]
        return (lits+clen)/(2.0+lits),i+lits+clen,code
    def en_long_copy(self,i,d,lits):
        found,clen = self.searchback(d,i,lits, max(0,i-(32768-lits)),i+lits,34)
        if found < 0:
            return -1,i,''
        index = -(found-i+1)+lits

        code = bytearray(3)
        bitstruct_pack(code, [
            (2,      [(2,0)             ]),
            (index,  [(6,16),(3,8),(6,2)]),
            (clen-3, [(5,11)            ]),
            (lits,   [(2,22)            ])])

        #check = ncbits_of(code, (6, 16),(3,8),(6,2))
        #if check != index:
        #    print "ERROR", index,check, clen-3, lits
        code += d[i:i+lits]
        return (lits+clen)/(3.0+lits),i+lits+clen,code
    def en_short_run(self,i,d):
        initial_color = d[i]
        end = i
        while end < len(d) and d[end] == initial_color:
            end += 1
        if end-i < 3:
            return -1,i,''
        runlength = min(18,end-i)
        code = bytearray(2)
        bits_pack(code, 0xE,           4, 0)
        bits_pack(code, runlength-3,   4, 4)
        bits_pack(code, initial_color, 8, 8)
        return (runlength/2.0, i+runlength, code)
    def en_long_run(self,i,d):
        initial_color = d[i]
        end = i
        while end < len(d) and d[end] == initial_color:
            end += 1
        if end-i < 3:
            return -1,i,''
        runlength = min(258,end-i)
        code = bytearray(3)
        bits_pack(code, 0xF0,          8, 0)
        bits_pack(code, runlength-3,   8, 8)
        bits_pack(code, initial_color, 8, 16)
        return (runlength/3.0, i+runlength, code)
    def condense_opcodes(self, codes):
        condensed = []
        data=bytearray()
        data_opcode = bytearray()
        for c in codes:
            if c[0]&0xF0 == 0xC0:
                data_opcode.extend(c[1:])
            else:
                if data_opcode: data += self.merge_C0(data_opcode)
                data += c
                data_opcode = bytearray()
        data += self.merge_C0(data_opcode)
        return data
    def merge_C0(self, ops):
        data = bytearray()
        assert not len(ops)%4
        while ops:
            chunk = min(64,len(ops))
            data += chr(0xC0 + chunk/4 - 1)
            data += ops[:chunk]
            ops = ops[chunk:]
        return data

class General(DelvImage):
    has_header = True

class TileSheet(DelvImage):
    canonical_size = 32,512

class Portrait(DelvImage):
    canonical_size = 64,64

class Landscape(DelvImage):
    canonical_size = 288,32


class SkillIcon(DelvImage):
    """Class for handling the small skill icons that are stored
       uncompressed as indexed color data.
       Methods with the same name generally work as in DelvImage."""
    canonical_size = 32,16
    width = 32
    height = 16
    logical_width = 32
    logical_height = 16
    #def __init__(self, data=None):
    #    self.image = data
    #    self.data = data
    #    if not self.image: self.image = bytearray(32*16)
    def decompress(self, src, *argv):
        self.image = src
        self.data = None
    def compress(self, *argv):
        self.data = self.image

_CLASS_HINTS = {142:General, 141:TileSheet, 135:Portrait, 131:Landscape,
                137:SkillIcon}
_NAME_HINTS = {'general':General,'tiles':TileSheet,'portrait':Portrait,
               'landscape':Landscape,'sprite':TileSheet,'sized':General,
               'icon':SkillIcon
               }



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
# Maps and prop lists. Convenience utilities for map visualization.
import delv.util
import store
import array

class PropListEntry(object):
    def __init__(self, flags, loc, aspect, proptype, d3, propref, storeref, u):
        self.flags = flags;self.loc=loc;self.aspect=aspect&0x1F
        self.rotated = aspect&0xE0
        self.proptype=proptype;self.d3=d3;self.propref=propref
        self.storeref=storeref;self.u=u
    def show_in_map(self):
        # This is probably wrong - many details yet to be determined. FIXME
        return not (self.flags & 0x4C)
    def get_loc(self): # FIXME
        return self.loc[0],self.loc[1]
    # Persistence data fields
    def get_d1(self):
        return self.d3>>8
    def set_d1(self,v):
        if v > 255: raise ValueError, "d1 is an 8-bit quantity"
        self.d3 = (self.d3&0x00FF)|(v<<8)
    def get_d2(self):
        return self.d3&0xFF
    def set_d2(self,v):
        if v > 255: raise ValueError, "d2 is an 8-bit quantity"
        self.d3 = (self.d3&0xFF00)|v
    def set_d3(self,v):
        self.d3 = v
    def get_d3(self):
        return self.d3
    def debug(self, library): 
        # Man I need to refactor, look at that nomenclature
        if self.show_in_map():
            t = library.get_tile(library.get_prop(
                self.proptype).get_tile(self.aspect))
            return "%03X:%X:%X-%08X[%08X]-%s(%04X)"%(
                self.proptype, self.aspect, 
                self.rotated, t.attributes, t.draw_priority(),
                t.get_name(), self.get_d3())
        else:
            return "{%04X:%02X(%04X)}"%(
                self.proptype | (self.aspect<<10) | (self.rotated<<5), 
                self.flags, self.get_d3())

class PropList(store.Store):
    def __init__(self, src):
        store.Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.props = []
        self.propsat = {}
    def load_from_bfile(self):
        while not self.src.eof():
            flags=self.src.read_uint8()
            loc= self.src.read_xy24()
            aspect,proptype = self.src.read_uint6_uint10()
            d3 = self.src.read_uint16()
            propref = self.src.read_uint32()
            storeref = self.src.read_uint16()
            u= self.src.read_uint16()
            nprop = PropListEntry(flags, loc,
                aspect, proptype,d3, propref, storeref,u)
            self.props.append(nprop)
            if not self.propsat.has_key(loc): self.propsat[loc] = []
            self.propsat[loc].append(nprop)
        #for prlist in self.propsat.values(): 
        #    prlist.sort(key=lambda p:p.draw_order())
    def props_at(self,loc):
        return self.propsat.get(loc,[])
    def draw_order(self):
        # NOTE should check to make sure this is always true... otherwise
        # this function needs to sort them as well.
        return self.props.__iter__()
    def __iter__(self):
        return self.props.__iter__()
    def __getitem__(self, n):
        return self.props[n]
    def __setitem__(self, k, v):
        self.props[k] = v
    def __len__(self):
        return len(self.props)

class Map(store.Store):
    def __init__(self, src):
        store.Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.rows = []
    def get_tile(self, x, y):
        """Return base tile at this location."""
        return self.map_data[x+y*self.width]
    def load_from_bfile(self):
        self.width = self.src.read_uint16()
        self.height = self.src.read_uint16()
        self.unknown = self.src.read_uint16()
        assert not self.unknown

        self.roof_layer_size = self.src.read_uint16()
        self.roof_underlayer_size = self.src.read_uint16()
        assert self.roof_layer_size == self.roof_underlayer_size
        
        # if 0, show blackness instead
        self.horizontal_edge_propagation = self.src.read_uint8()
        self.vertical_edge_propagation = self.src.read_uint8()

        # which zoneport you end up at when you walk off of the map in
        # the four cardinal directions. Usually the same, but can be 
        # different - see 0x8026 for how the Cademia bridge is implemented
        self.exit_zoneport_north = self.src.read_uint16()
        self.exit_zoneport_east = self.src.read_uint16()
        self.exit_zoneport_south = self.src.read_uint16()
        self.exit_zoneport_west = self.src.read_uint16()

        # These seem to be zero. But if not, let's draw attention to it
        self.padding = self.src.readb(12)
        for b in self.padding: assert not b


        roof_length = 0x40*self.roof_layer_size+0x40*self.roof_underlayer_size
        self.roof_data = array.array('H')
        for _ in xrange(roof_length/2): 
            self.roof_data.append(self.src.read_uint16())
        self.map_data = array.array('H')
        for _ in xrange(self.width*self.height):
            self.map_data.append(self.src.read_uint16())
    

class Level(object):
    pass



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import store


class Tile(object):
    def __init__(self, index, namecode, attributes, fauxprop, image):
        self.index=index
        self.namecode = namecode
        self.attributes = attributes
        self.fauxprop,self.fauxprop_aspect,self.fauxprop_rotate = fauxprop
        self.mask = None
        self.rotated_mask = None
        self.image = str(image)
        self.rotated_cache = None
        self.requires_mask =(attributes & 0xFF000000) and '\x00' in self.image
    def draw_priority(self):
        if not self.attributes: return 0xFFFFFFFF
        return (self.attributes&0xBFCDFD1C)
    def get_name(self,plural=False):
        return store.namecode(self.namecode, plural)
    def get_image(self, rotated=False): 
        if rotated:
            if not self.rotated_cache: 
                self.rotated_cache = self.rotate()
            return self.rotated_cache
        else:
            return self.image
    def rotate(self):
        rotated = bytearray(32*32)
        for y in xrange(32):
            rotated[y*32:y*32+32] = self.image[y::32]
        return str(rotated)
    def get_pixmap_mask(self,gtk,rotated=False):
        """This is really just to save redelv the bother of having 
           a separate cache of tile masks..."""
        if self.mask and not rotated: return self.mask
        if self.rotated_mask and rotated: return self.rotated_mask
        image = self.get_image(rotated)
        mask = gtk.gdk.Pixmap(None, 32, 32, 1)
        on = mask.new_gc(foreground=gtk.gdk.Color(pixel=1),
             function=gtk.gdk.COPY)
        off = mask.new_gc(foreground=gtk.gdk.Color(pixel=0),
             function=gtk.gdk.COPY) 
        mask.draw_rectangle(off, True, 0,0,32,32)
        for n,pixel in enumerate(image):
            if pixel != '\x00': 
                mask.draw_point(on, n%32, n//32)
        #    else:
        #        self.mask.draw_point(off, n%32, n//32)
        if rotated:
            self.rotated_mask = mask
        else:
            self.mask = mask
        return self.rotated_mask if rotated else self.mask

class CompoundTile(Tile):
    def __init__(self, index, namecode, attributes, fauxprop, library, 
        composition):
        image = bytearray(32*32)
        for n,(resid, tile, segment) in enumerate(composition):
            chunk = library.get_object(resid).get_subtile(tile,segment)
            i = (n%4)*8 + (n//4)*8*32
            for r in xrange(8):
                 image[i+32*r:i+32*r+8] = chunk[8*r:8*r+8]
        Tile.__init__(self, index, namecode, attributes, fauxprop, image)



#!/usr/bin/env python
# Copyright 2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import archive,hints,tile,prop
import array
class Library(object):
    """This class is a wrapper around one or more Delver archives, and 
       facilitates retrieving various kinds of objects by their resource
       ID. I.e, whereas the delv.archive.Scenario object will return a 
       delv.archive.Resource object when asked for resource ID 0x8E91,
       the library will return a delv.graphics.TileSheet object. 

       The Library also provides facilities that integrate multiple
       resources into logical parts, e.g. it unifies graphical tiles
       with their names and properties, and allows them to be retrieved
       under a simple API.
 
       The library is a read-only facility and it assumes that its underlying
       archive does not change from under it. If that assumption is not true,
       i.e. if you are using it as part of an editor, you will need to tell
       the library the underlying archive has changed with .update(). This 
       will purge all caches."""
    def __init__(self, *archives):
        """Create a new library, using the archives provided. The search path
           will be from last to first, i.e. if you say Library(A, B, C) and 
           archives B and C both contain a resource 0xFFFF, and you ask for
           0xFFFF, you will get the one from archive C."""
        self.cache = {}
        self.archives = archives[::-1]
        self.load_tiles()
        self.load_props()

    def load_props(self):
        self.props = []
        proptiles = self.get_object(0xF000)
        xoffsets = self.get_object(0xF011)
        yoffsets = self.get_object(0xF012)
        for pid,tile in enumerate(proptiles):
            self.props.append(prop.Prop(pid, tile, zip(
                       xoffsets[pid<<5:(pid<<5)+32],
                       yoffsets[pid<<5:(pid<<5)+32]), scripts=None))

    def load_tiles(self):
        tilenames = self.get_object(0xF004)
        tileattrs = self.get_object(0xF002)
        tilecompositions = self.get_object(0xF013)
        tilefauxprops = self.get_object(0xF010)
        tilesheets = self.objects(hints.GRAPHICS_TILESHEET, True)
        self.tiles = []
        tile_Nothing = tile.Tile(0,tilenames[0], tileattrs[0],
            tilefauxprops[0],
            tilesheets[0].get_tile(0))

        
        for n,attr in enumerate(tileattrs):
            tileres_n = (n >> 4)&0xFF
            if not (tileattrs[n] or tilesheets[tileres_n]): 
                self.tiles.append(tile_Nothing)
                continue
            if n < 0x1000:
                self.tiles.append(tile.Tile(n,tilenames[n], tileattrs[n],
                    tilefauxprops[n],
                    tilesheets[tileres_n].get_tile(n&0x00F)))
            elif n < 0x2000:
                self.tiles.append(tile.CompoundTile(n,tilenames[n],
                    tileattrs[n],tilefauxprops[n],
                    self,tilecompositions[n-0x1000]))
            else: 
                assert 0, "Turns out there are more tiles than I thought."
                
    def get_tile(self, tid):
        """Returns the specified Tile, or "Nothing" if there is none such."""
        return self.tiles[tid]
    def get_prop(self, ptype):
        return self.props[ptype]

    def objects(self,si=None,dense=False):
        if dense:
            return [self.get_object((si,n)) for n in xrange(256)]
        else:
            return [self.get_object(resid) for resid in self.resource_ids(si)]
    def resources(self,si=None):
        return [self.get_resource(resid) for resid in self.resource_ids(si)]
    def resource_ids(self, si=None):
        ids = set()
        for archive in self.archives:
            ids.update(archive.resource_ids(si))
        return list(ids)
    def get_resource(self, resid):
        """Get a resource by its resource ID or subindex,n pair."""
        for archive in self.archives:
            r = archive.get(resid)
            if r: return r
        return None
    def get_object(self, resid):
        """Get the appropriate kind of object for a specified resource."""
        r = self.get_resource(resid)
        if not r: return None
        if r in self.cache: return self.cache[r]
        C = hints.class_by_resid(r.resid)
        if not C: return None
        ob = C(r)
        self.cache[r] = ob
        return ob
    



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import graphics, sound, store, level
GRAPHICS_LANDSCAPE = 131
GRAPHICS_PORTRAIT = 135
GRAPHICS_TILESHEET = 141
_OBJECT_SI_HINTS = {
   127: level.Map,
   128: level.PropList,
   131: graphics.Landscape, 
   135: graphics.Portrait,
   137: graphics.SkillIcon,
   141: graphics.TileSheet,
   142: graphics.General,
   144: sound.Asnd,
}
_OBJECT_RESID_HINTS = {
  0xF004: store.TileNameList,
  0xF000: store.PropTileList,
  0xF010: store.TileFauxPropsList,
  0xF011: store.ByteList,
  0xF012: store.ByteList,
  0xF002: store.TileAttributesList, 
  0xF013: store.TileCompositionList,
  0x8EFF: graphics.General,
}
def class_by_resid(resid):
    """Try to get a class that can handle the resource provided,
       otherwise return None"""
    return _OBJECT_RESID_HINTS.get(resid,
           _OBJECT_SI_HINTS.get((resid >> 8) - 1, None))
    


_RES_HINTS = {
 0x0101: "Global Symbol List",

 0x0201: "Character Names",
 0x0203: "Character Class Names",
 0x0204: "Character Class Descriptions",
 0x0205: "Character Class Stats",
 0x0206: "Character Class Skills",
 0x0218: "Sign Text",
 0x0219: "Scroll Text",
 0x021A: "Quest Text",
 0x021B: "Book Item Text",
 0x021D: "Bookshelf Text",
 0x021F: "Ring Inscriptions",
 0x0220: "Gravestone Inscriptions",
 0x0241: "Death Text",
 0x0242: "Defeat Text",
 0x0243: "Victory Text",

 0x0410: "Attack Nearest",
 0x0411: "Attack Weakest",
 0x0413: "Defend",
 0x0414: "Berserk",
 0x0415: "Missile Script",
 0xF008: "Monster Statistics",
 0xF000: "Prop-Tile Associations",
 0xF004: "Tile Names",
 0xF002: "Tile Attributes",
 0xF013: "Composed Tiles",
 0xF010: "Faux Prop Information",
 0xF011: "Prop-aspect X offsets",
 0xF012: "Prop-aspect Y offsets",
 0xF015: "Peristence Store Symbols"
}
_SCEN_HINTS = {
    1: "String Lists",2: "Static Data",3: "AI Scripts",4: "Script Data",
    7: "Shared Dialogue Scripts",
    8: "Scripts", 9: "Scripts", 10: "Scripts", 11: "Scripts", 12: "Scripts",
    13: "Scripts", 14: "Scripts", 16: "Object Scripts", 19: "Zone Scripts",
    15: "Dialogue Scripts", 20: "Composite Zone Scripts", 
    23: "Dialogue Scripts", 25: "Skill and Action Scripts", 
    26: "Area Scripts", 27: "Area Scripts", 47: "Character Action Scripts",
    127: "Maps", 128: "Prop Lists", 131: "Landscape Graphics", 
    135: "Character Portraits", 137: "Skill Icons", 141: "Tile Graphics",
    142: "General Graphics", 143: "Music", 144: "Sounds", 
    239: "Script Data", 254: "Patch Description",
    187: "Metadata from delv",
    129: "Explored Area Bitmaps", 223: "Journal Entries"
}




#!/usr/bin/env python
# Copyright 2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import store

class Prop(object):
    def __init__(self, pid, tile, offset_by_aspect, scripts):
        self.scripts=scripts
        self.tile = tile
        self.prop_type = pid
        self.offset_by_aspect = offset_by_aspect
    def get_tile(self, aspect, rotated=False):
        return self.tile + (aspect)
    def get_debug_tile(self, aspect):
        return 0x017F
    def get_offset(self, aspect, rotated=False):
        return self.offset_by_aspect[aspect]




#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
# This file addresses sundry storage types used within Delver Archives,
# and as such is mostly a helper for other parts of delv. 

# FIXME this file has gotten seriously out of hand with copy-pasted junk;
# it needs a more object-oriented refactoring


import util, archive
import cStringIO as StringIO
import array, bisect
import tile

class Store(object):
    def __init__(self, src):
        self.data = None
        self.set_source(src)
    def set_source(self, src):
        self.src = None
        if src and issubclass(src.__class__, util.BinaryHandler):
            self.src = src
            self.res = None
        elif src and issubclass(src.__class__, archive.Resource):
            self.src = src.as_file()
            self.res = src
        elif src:
            self.src = util.BinaryHandler(src)
            self.res = None
    def get_data(self):
        if not self.data:
            buf = StringIO.StringIO()
            bh = util.BinaryHandler(buf)
            self.write_to_bfile(bh)
            # I wonder why StringIO doesn't have a method that does this:
            self.data = bytearray(buf.getvalue())
        return self.data 

class SymbolList(Store):
    pass

class TileAttributesList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.contents = array.array("L")
    def load_from_bfile(self):
        while not self.src.eof():
            self.contents.append(self.src.read_uint32())
    def __iter__(self): return self.contents.__iter__()
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        for attr in self.contents: dest.write_uint32(attr)
    def __getitem__(self, n):
        return self.contents[n]
    def __setitem__(self, n, value):
        self.contents[n] = value
    def __len__(self): return len(self.contents)


class ByteList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.contents = bytearray()
    def load_from_bfile(self):
        self.contents = self.src.readb()
    def __iter__(self): return self.contents.__iter__()
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        dest.write(self.contents)
    def __getitem__(self, n):
        return self.contents[n]
    def __setitem__(self, n, value):
        self.contents[n] = value
    def __len__(self): return len(self.contents)

class PropTileList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.contents = array.array("H")
    def load_from_bfile(self):
        while not self.src.eof():
            self.contents.append(self.src.read_uint16())
    def __iter__(self): return self.contents.__iter__()
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        for attr in self.contents: dest.write_uint16(attr)
    def __getitem__(self, n):
        return self.contents[n]
    def __setitem__(self, n, value):
        self.contents[n] = value
    def __len__(self): return len(self.contents)

class TileFauxPropsList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.contents = []
    def load_from_bfile(self):
        while not self.src.eof():
            word = self.src.read_uint16()
            self.contents.append((word&0x3FF, (word>>10)&0x1F, word>>15))
    def __iter__(self): return self.contents.__iter__()
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        for ptype,aspect,rotate in self.contents: dest.write_uint16(
            ptype|(aspect<<10)|(rotate<<15))
    def __getitem__(self, n):
        return self.contents[n]
    def __setitem__(self, n, value):
        self.contents[n] = value
    def __len__(self): return len(self.contents)

# maybe store should subclass list?
class TileCompositionList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
        self.empty()
        if self.src: self.load_from_bfile()
    def empty(self):
        self.contents = []
    def load_from_bfile(self):
        while not self.src.eof():
            self.contents.append(self.parse([
                self.src.read_uint16() for _ in xrange(0x10)]))
    def parse(self, tw):
        return ((0x8E00|((t>>4)&0x00FF),t&0x000F,(t&0xF000)>>12) for t in tw)

    def __iter__(self): return self.contents.__iter__()
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        for attr in self.contents: dest.write_uint32(attr)
    def __getitem__(self, n):
        return self.contents[n]
    def __setitem__(self, n, value):
        self.contents[n] = value
    def __len__(self): return len(self.contents)


def namecode(name, plural):
    if '\\' in name:
        stem = name[:name.find('\\')]
        ending = name[name.find('\\')+1:]
        if '/' in ending:
            return stem+ending.split('/')[1 if plural else 0]
        else:
            return stem+ending if plural else stem
    else:
        return name


class TileNameList(Store):
    def __init__(self, src):
        Store.__init__(self, src)
       
        self.empty()
        if self.src: self.load_from_bfile()
    def items(self):
        """Return a list of cutoff,name pairs."""
        return zip(self.cutoffs,self.names)
    def __iter__(self): return self.cutoffs.__iter__()
    def empty(self):
        """Purge the contents of the TileNameList."""
        self.names = []
        self.cutoffs = array.array('H')
    def write_to_bfile(self, dest):
        if dest is None: dest = self.src
        dest.seek(0)
        for cutoff, name in zip(self.cutoffs,self.names):
            dest.write_uint16(cutoff)
            dest.write_cstring(name)
    def load_from_bfile(self):
        while not self.src.eof():
            value = self.src.read_uint16()
            name = self.src.read_cstring()
            if not name and not value: break
            self.cutoffs.append(value)
            self.names.append(name)
    def namecode(self, name, plural):
        return namecode(name,plural)
    def get_name(self, idx, plural=False):
        return self.namecode(self[idx], plural)
    def __getitem__(self, idx):
        return self.names[bisect.bisect_left(self.cutoffs,idx)]
    def __setitem__(self, idx, val):
        "Note that setitem will not create new pairs, it just sets names."
        self.names[bisect.bisect_left(self.cutoffs,idx)] = val
    def __delitem__(self, cutoff):
        rmidx = bisect.bisect_left(self.cutoffs,cutoff)
        del self.cutoffs[rmidx]
        del self.names[rmidx]
    def append(self, cutoff, name):
        "Insert the name and cutoff into this object."
        insertidx = bisect.bisect_left(self.cutoffs,cutoff)
        self.cutoffs.insert(insertidx, cutoff)
        self.names.insert(insertidx, name)
    def extend(self, data):
        "Data - a list of (cutoff,name) pairs, or a dict of {cutoff:name}."
        if isinstance(data,dict):
             data = data.items()
        for pair in data: self.append(*pair)
    def __len__(self):
        return len(self.names)



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Version: 0.20
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
import cStringIO as StringIO, struct

# This whole file is intended to be 'private' to delv; it isn't part of the 
# public API.

def int_to_bits(value,size):
    result = bytearray(size)
    j =0 
    for i in xrange(size-1,-1,-1):
        result[j] = (value >> i)&1
        j += 1
    return result

def bytes_to_bits(src):
    src = bytearray(src) # Note that if src is an integer, a new empty bytearray is made.
    # this is not efficient, but it does at least make sense semantically.
    result = bytearray(len(src)*8)
    ri = 0
    for byte in src:
        for bi in xrange(7,-1,-1): 
            result[ri] = (byte>>bi)&1
            ri += 1
    return result

def bits_to_bytes(src):
    result = bytearray((len(src)+7)//8)
    byi = 0
    bi = 0
    for n,bit in enumerate(src):
        byte_index = n//8
        bit_index = n % 8
        result[byte_index] |= bit << (7-bit_index)
    return result

def bitstruct_pack(target, pairs):
    b = bytes_to_bits(target)
    for value, fieldspec in pairs:
        for size, index in fieldspec[::-1]:
            fieldbits = value & (0xFFFFFFFFFFFFFFFFL >> (64-size))
            b[index:index+size] = int_to_bits(value,size)
            value >>= size
    t =  bits_to_bytes(b)
    for x in xrange(len(t)): target[x]=t[x]


def bits_pack(target, value, size, index,debug=False):
    """Alter bytearray target so that size bits of target starting
       at index are replaced by value."""
    bit_index = index % 8
    startbyte = index // 8
    endbyte = (index+size+7) // 8
    section = bytes_to_bits(target[startbyte:endbyte])
    #if debug:print "BEFORE", ''.join(["%01d"%b for b in section])
    section[bit_index:bit_index+size] = int_to_bits(value,size)
    #if debug:print "AFTER ", ''.join(["%01d"%b for b in section])
    target[startbyte:endbyte] = bits_to_bytes(section)

def ncbits_pack(target, value, *fields):
    """Alter target to contain the value given, broken up into
       nonconsecutive bitfields using the same sematics as ncbits_of.
       
       Look on this, ye coders, and repeat the sacred mantra:
          "Premature optimization is the root of all evil."
    """
    # This would be more efficient if we made it only convert to
    # bits and back once, or if we rewrote it to be like bits_of (I
    # actually wrote that first before I thought of this simpler way)
    # but why bother, it's only used compressing graphics, which is an
    # infrequent operation (whereas decompression happens a lot.)
    for size,index in fields[::-1]:
        fieldbits = value & (0xFFFFFFFFFFFFFFFFL >> (64-size))
        bits_pack(target, fieldbits, size, index)
        value >>= size
    

def ncbits_of(data, *fields):
    """Returns an integer bit field from the bytearray data, 
       the integer being made up of all the fields combined.
       Each field is a tuple in the format (size,index). 
       e.g. if  you had a bytearray foo of three bytes as follows:
       
         -----AB-, ---CDEFG, HI--JKL-   

       And you wanted to extract the bits ABCDEFGHIJKL as one
       12-bit integer, you'd say:
         ncbits_of(foo, (2,5), (7,11), (3,20))

       Note that if you ever use this function other than to be
       compatible with something already using nonconsecutive bits,
       I am pretty sure that D. E. Knuth will come to your home and
       perform an exorcism involving a shotgun and rock salt. Or at least,
       he ought to.
    """
    result = 0
    position = 0
    for size,index in fields:
        result <<= size
        result |= bits_of(data, size, index)
    return result

def bits_of(data, size, index):
    "Read a slice of the bytearray data bitwise - indices in bits"
    byte_index = index // 8
    bit_index = index % 8
    byte_end = (size+index) // 8 
    bit_size = size % 8 
    result = data[byte_index] & (0xFF >> bit_index)
    while byte_index < byte_end:
        byte_index += 1
        if len(data) == byte_index: 
            result <<= 8; break
        result = (result << 8) | data[byte_index]
    result >>= 8 - (bit_index + bit_size) % 8
    return result
        


class BinaryHandler(object):
    S_offlen = struct.Struct('>LL')
    S_uint32 = struct.Struct('>L')
    S_sint32 = struct.Struct('>l')
    S_uint8 = struct.Struct('B')
    S_uint16 = struct.Struct('>H')
    S_sint16 = struct.Struct('>h')

    # Wishing for a more elegant alternative
    def eof(self):
        rv = self.file.read(1)
        self.file.seek(-1, 1)
        return rv is ''
    def seek(self, *vargs, **kwargs):
        self.file.seek(*vargs, **kwargs)
    def read(self, *vargs, **kwargs):
        return self.file.read(*vargs, **kwargs)
    def readb(self, *vargs, **kwargs):
        return bytearray(self.file.read(*vargs, **kwargs))
    def write(self, *vargs, **kwargs):
        self.file.write(*vargs, **kwargs)
    def tell(self, *vargs, **kwargs):
        return self.file.tell(*vargs, **kwargs)
    def truncate(self, *vargs, **kwargs):
        return self.file.truncate(*vargs,**kwargs)

    def __init__(self, file):
        if hasattr(file, 'read') and hasattr(file, 'write'):
            self.file = file
        elif hasattr(file, '__getitem__'):
            self.file = StringIO.StringIO(file)
    def write_struct(self, s, v, offset=None):
        if offset is not None: self.seek(offset)
        if type(v) is int:
            self.write(s.pack(v)) 
        else:
            self.write(s.pack(*v))
    def write_uint8(self, v, offset=None):
        self.write_struct(self.S_uint8, v, offset)
    def write_uint16(self,v,offset=None):
        self.write_struct(self.S_uint16, v, offset)
    def write_uint6_uint10(self,v,offset=None):
        self.write_uint16((v[0]<<10)|v[1],offset)
    def write_sint16(self,v,offset=None):
        self.write_struct(self.S_sint16, v, offset)
    def write_uint32(self,v,offset=None):
        self.write_struct(self.S_uint32, v, offset)
    def write_offlen(self,offs,length,offset=None):
        self.write_struct(self.S_offlen, (offs,length), offset)

    def write_sint24(self,v,offset=None):
        if offset is not None: self.seek(offset)
        self.write_uint8((v&0xFF0000)>>16)
        return self.write_uint16((v&0x00FFFF))

    def write_xy24(self,x,y,offset=None):
        if offset is not None: self.seek(offset)
        self.write_uint8((x&0xFF0)>>4)
        self.write_uint8(((x&0x00F)<<4)&((y&0xF00)>>8))
        return self.write_uint8(y&0x0FF)

    def write_vm32(self, flags, v, offset=None):
        if offset is not None: self.seek(offset)
        self.write_uint8(flags)
        return self.write_sint24(v)

    def write_pstring(self, s, offset=None):
        if offset is not None: self.seek(offset)
        "Write a pascal string - 8 bit length followed by data."
        assert len(s) < 256
        self.write_uint8(len(s))
        return self.write(s)
    def write_str31(self, s, offset=None):
        if offset is not None: self.seek(offset)
        assert len(s) < 32
        self.write_uint8(len(s))
        self.write('\x00'%(31-len(s)))
    def write_cstring(self, s, offset=None):
        "Write a null-terminated string."
        if offset is not None: self.seek(offset)
        self.write(s)
        return self.write('\x00')
    def write_fixed16(self, s, offset=None):
        "Write 8.8 fixed-point number."
        if offset is not None: self.seek(offset)
        self.write_uint8(int(s))
        self.write_uint8(int(round(255*(s-int(s)))))

    def read_struct(self, s, offset=None):
        if offset is not None: self.seek(offset)
        return s.unpack(self.read(s.size))
    def read_offlen(self, offset=None):
        return self.read_struct(self.S_offlen, offset)
    def read_uint8(self, offset=None):
        "Read 8-bit unsigned integer"
        return self.read_struct(self.S_uint8, offset)[0]
    def read_uint16(self, offset=None):
        "Read 16-bit big endian unsigned integer."
        return self.read_struct(self.S_uint16, offset)[0]
    def read_uint6_uint10(self,offset=None):
        v = self.read_uint16(offset)
        return v>>10, v&0x3FF
    def read_sint16(self, offset=None):
        "Read 16-bit big endian signed integer."
        return self.read_struct(self.S_sint16, offset)[0]
    def read_sint24(self, offset=None):
        "Return a signed 24-bit integer."
        uvar = (self.read_uint8(offset)<<8) | self.read_uint16()
        if uvar & 0x800000:
            uvar = ~uvar + 1
        return uvar
    def read_xy24(self, offset=None):
        "Read packed 12-bit xy coordinates, as used in prop lists."
        if offset is not None: self.seek(offset)
        d = self.readb(3)
        return (d[0]<<4)|(d[1]>>4), ((d[1]&0x0F)<<4)|d[2]
    def read_uint32(self, offset=None):
        "Read 32-bit unsigned integer."
        return self.read_struct(self.S_uint32, offset)[0]
    def read_vm32(self, offset=None):
        "Read 24-bit signed integer and 8-bit flags. (Flags returned first.)"
        return self.read_uint8(offset), self.read_sint24()
    def read_pstring(self, offset=None):
        "Read a Pascal String (Length byte followed by that many data bytes)"
        size = self.read_uint8(offset)
        return self.read(size)
    def read_str31(self, offset=None):
        "Read a Str31."
        size = self.read_uint8(offset)
        return self.read(31)[:size]
    def read_fixed16(self,offset=None):
        "Read an 8.8 Fixed number."
        units = self.read_uint8(offset)
        fraction = self.read_uint8()
        return units + fraction/256.0
        
    def read_cstring(self, offset=None):
        if offset is not None: self.seek(offset)
        buf = bytearray()
        while True:
            b = self.read(1)
            if b == '\0' or not b: break
            buf += b
        return str(buf)

class UnimplementedFeature (Exception): pass



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 
# Maps and prop lists. Convenience utilities for map visualization.
import util, archive, store
from util import bitstruct_pack, bits_pack,bits_of
import array
import cStringIO as StringIO

class SoundError(Exception): pass

class Sound(store.Store):
    pass

class Asnd(Sound):
    """Class for Delver sound resources. (Might be 'asnd' for "Ambrosia 
       Sound Tool" but that's speculative.)"""
    def __init__(self, src=None):
        self.samples = array.array('h')
        self.rate = 22050
        self.duration =0 
        self.data = None
        self.set_source(src)
        if self.src: self.load_from_bfile()
        
    def write_to_bfile(self, dest=None):
        if dest is None: dest = self.src
        dest.seek(0)
        dest.write('asnd')
        dest.write_uint32(self.duration)
        dest.write_uint16(self.rate)
        dest.write_uint16(self.flags)
        for sample in self.samples:
            # Yes, this seems stupid, but it's the way it is
            # No idea why 8 bit sound is stored in 16 bit shorts.
            dest.write_sint16(sample>>8)
        
    def load_from_bfile(self):
        self.src.seek(0)
        if self.src.read(4) != 'asnd': raise SoundError, "Bad magic number"
        self.duration = self.src.read_uint32()
        self.rate = self.src.read_uint16()
        self.flags = self.src.read_uint16()
        # decode pcm
        samp = 0
        while not self.src.eof():
            #samp += self.src.read_sint16()
            #self.samples.append(samp) 
            self.samples.append(self.src.read_sint16()<<8)
        self.data = None
    def get_rate(self):
        return self.rate
    def get_samples(self):
        "Get an array of 16-bit signed integers."
        return self.samples
    def set_samples(self, newsamples):
        self.samples = array.array('h',newsamples)
        self.data = None
        # It's actually size, not duration... anyway, may need to pad.
        
        if len(self.samples) < 512:
            self.samples.extend([0]*(512-len(self.samples)))
            self.duration = 0
        else:
            self.samples.extend([0]*(1024 - len(self.samples)%1024))
            self.duration = (len(self.samples)-512)/1024
    def set_rate(self, newrate):
        self.rate = newrate
        self.data = None
        
    
class SoundSND(Sound):
    pass


class MusicError(Exception): pass

MUSIC_COMPONENT_TYPE = 'musi'
SOFT_SYNTH_COMPONENT_SUBTYPE = 'ss  '
GM_SYNTH_COMPONENT_SUBTYPE = 'gm  '
NOTE_REQUEST = 1
PART_KEY = 4
MIDI_CHANNEL = 8
USED_NOTES = 11
GENERAL_EVENT = 0xF
INSTRUMENT_NAMES = """Acoustic Grand Piano
Bright Acoustic Piano 
Electric Grand Piano
Honky-tonk Piano 
Electric Piano 1 
Electric Piano 2 
Harpsichord 
Clavi 
Celesta 
Glockenspiel 
Music Box 
Vibraphone 
Marimba 
Xylophone 
Tubular Bells 
Dulcimer 
Drawbar Organ 
Percussive Organ 
Rock Organ 
Church Organ 
Reed Organ 
Accordion 
Harmonica 
Tango Accordion 
Acoustic Guitar (nylon) 
Acoustic Guitar (steel) 
Electric Guitar (jazz) 
Electric Guitar (clean) 
Electric Guitar (muted) 
Overdriven Guitar 
Distortion Guitar 
Guitar harmonics 
Acoustic Bass 
Electric Bass (finger) 
Electric Bass (pick) 
Fretless Bass 
Slap Bass 1 
Slap Bass 2 
Synth Bass 1 
Synth Bass 2 
Violin 
Viola
Cello 
Contrabass 
Tremolo Strings 
Pizzicato Strings 
Orchestral Harp 
Timpani 
String Ensemble 1 
String Ensemble 2 
SynthStrings 1 
SynthStrings 2 
Choir Aahs 
Voice Oohs 
Synth Voice 
Orchestra Hit 
Trumpet 
Trombone 
Tuba 
Muted Trumpet 
French Horn 
Brass Section 
SynthBrass 1 
SynthBrass 2 
Soprano Sax
Alto Sax 
Tenor Sax 
Baritone Sax 
Oboe 
English Horn
Bassoon 
Clarinet 
Piccolo
Flute 
Recorder 
Pan Flute
Blown Bottle
Shakuhachi 
Whistle 
Ocarina 
Lead 1 (square) 
Lead 2 (sawtooth)
Lead 3 (calliope) 
Lead 4 (chiff) 
Lead 5 (charang) 
Lead 6 (voice) 
Lead 7 (fifths) 
Lead 8 (bass + lead)
Pad 1 (new age)
Pad 2 (warm) 
Pad 3 (polysynth)
Pad 4 (choir) 
Pad 5 (bowed) 
Pad 6 (metallic) 
Pad 7 (halo) 
Pad 8 (sweep) 
FX 1 (rain) 
FX 2 (soundtrack) 
FX 3 (crystal) 
FX 4 (atmosphere) 
FX 5 (brightness) 
FX 6 (goblins) 
FX 7 (echoes) 
FX 8 (sci-fi) 
Sitar 
Banjo 
Shamisen 
Koto 
Kalimba 
Bag pipe 
Fiddle 
Shanai 
Tinkle Bell 
Agogo 
Steel Drums 
Woodblock 
Taiko Drum 
Melodic Tom 
Synth Drum 
Reverse Cymbal 
Guitar Fret Noise 
Breath Noise 
Seashore 
Bird Tweet 
Telephone Ring 
Helicopter 
Applause 
Gunshot""".strip().split('\n')

class Music(Sound):
    SxCD = 1
    TxCD = 0
    def __init__(self, src=None):
        self.parts = {}
        self.qtma_commands = []
        self.midi_commands = []
        self.flags = [0,1,1] # no idea what these do, if anything
        if src: self.load_from_file(src)

    def load_from_file(self, src):
        self.src = util.BinaryHandler(src)
        body_offset = self.src.read_uint32()
        musi = self.src.read(4)
        if musi != MUSIC_COMPONENT_TYPE: 
            raise ValueError, "%s is not a music resource"%src
        self.flags = [self.src.read_uint32() for _ in xrange(3)]
 
        while self.src.tell() < body_offset:
            command = self.src.readb(4)
            operation = bits_of(command, 4, 0)
            if operation == GENERAL_EVENT:
                part = bits_of(command, 12, 4)
                event_length = bits_of(command, 16, 16)
                data_len = (event_length-2)*4
                data_offs = self.src.tell()
                self.src.seek(data_offs+data_len)
                footer = self.src.readb(4)
                subtype = bits_of(footer, 14, 2)
                event_length2 = bits_of(footer, 16,16)
                if event_length != event_length2:
                    raise MusicError, "Failed QTMA validity check."
                #print "General", part, event_length, subtype
                if subtype == MIDI_CHANNEL or subtype == USED_NOTES:
                    continue # we don't care
                elif subtype == NOTE_REQUEST:
                    next_op = self.src.tell()
                    self.src.seek(data_offs)
                    nrflags = self.src.read_uint8()
                    reserved = self.src.read_uint8()
                    polyphony = self.src.read_uint16()
                    typical = self.src.read_fixed16()
                    _align = self.src.read_uint16()
                    synthesizer_type = self.src.read(4)
                    synthesizer_name = self.src.read_str31()
                    instrument_name = self.src.read_str31()
                    instrument_number = self.src.read_uint32()
                    gm_number = self.src.read_uint32()
                    #print "\tNote request", nrflags, reserved,polyphony,
                    #print typical,repr(synthesizer_type),synthesizer_name,
                    #print instrument_name,instrument_number,gm_number
                    self.parts[part] = gm_number
                    self.src.seek(next_op)
                else:
                    raise MusicError, "Unknown QTMA general subtype %d"%subtype
            elif operation == 0x6: # End (marker event)
                #print "End    ", self.src.tell()
                break
            else:
                raise MusicError, "Unrecognized op, 0x%X@%d"%(
                    operation, self.src.tell()-4)
        self.src.seek(body_offset)
        self.load_qtma()
    def setup_pygame_midi_output(self, output):
        "Set up the MIDI output to play this music."
        for partnumber,instrument in self.parts.items():
            if instrument > 0x80: instrument &= 0x7F 
            output.set_instrument(instrument,partnumber)
    def load_qtma(self):
        # Load the body commands
        while not self.src.eof():
            command = self.src.readb(4)
            if   bits_of(command, 3, 0) == 1: # note event
                part = bits_of(command, 5, 3)
                pitch = bits_of(command, 6,8) + 32
                velocity = bits_of(command, 7, 14)
                duration = bits_of(command, 11,21)
                print "note", part, pitch,velocity,duration
            #elif bits_of(command, 4, 0) == 9: # extended note
            #    tail = self.src.readb(4)
            #    print "enote"
            elif bits_of(command, 3, 0) == 0: # rest 
                duration = bits_of(command, 24, 8)
                print "rest", duration
            elif bits_of(command, 4, 0) == 0xF: # general
                event_length = bits_of(command, 16, 16)
                data_len = (event_length-2)*4
                self.src.seek(data_len,1)
            #else:
            #    print "Unknown",("%02X "*4)%tuple(command), self.src.tell()


        
    def part(self,partnumber=None):
        """Return the general midi instrument associated with a part number."""
        return self.parts if partnumber is None else self.parts[partnumber]
    def set_midi(self, new):
        self.midi_commands = new
        self.qtma_commands = []
    def set_qtma(self, new):
        self.qtma_commands = new
        self.midi_commands = []
    def get_midi(self):
        "Return list of MIDI commands - in the format pygame.midi.write wants."
        if not self.midi_commands: self.convert_to_midi()
        return self.midi_commands
    def get_qtma(self):
        """Return list of QTMA commands."""
        if not self.qtma_commands: self.convert_to_qtma()
        return self.qtma_commands
    # private methods
    def convert_to_midi(self):
        pass
    def convert_to_qtma(self):
        pass



#!/usr/bin/env python
# Copyright 2014-2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Please do not make trouble for me or the Technical Documentation Project by
# using this software to create versions of the "Cythera Data" file which 
# have bypassed registration checks.
# Also, remember that the "Cythera Data" file is copyrighted by Ambrosia and
# /or Glenn Andreas, and publishing modified versions without their permission
# would violate that copyright. 
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 

import os
import util
#import numpy as np
import operator
import json, string, cStringIO as StringIO
from hints import _RES_HINTS, _SCEN_HINTS

def decrypt(data, prokey):
    """Decrypt the data provided with the given pro-key. The prokey is
       used to generate the seed value and parameters for the 
       pseudorandom number generator that is used to create the key."""
    cleartext = bytearray(len(data))
    key = prokey ^ (prokey>>8)
    m = ((prokey & 0x3f) <<2) + 1
    b = prokey >> 6

    for i in xrange(len(data)):
        key = (key*m + b) & 0xFFFF
        cleartext[i] = (data[i]^key)&0xFF

    return cleartext

encrypt = decrypt # Symmetric.

def entropy(data):
    """Return a statistical measure of the entropy of the given data.
       Encrypted (or compressed) data has high entropy. Note that this is
       not infallable and if the cleartext has very high entropy, e.g. in
       the case of compressed data, it may not be able to distinguish
       a successful decryption, at least in principle."""
    counts = [0]*256
    #counts = np.zeros(256) # Uncomment for faster entropy calculations
    # in exchange for a numpy dependency
    for c in data: counts[c] += 1
    base = len(data)/256.0
    for i in xrange(256):
         counts[i] -= base
         counts[i] *= counts[i]
    #counts -= len(data)/256.0
    #counts *= counts
    #return 1 - ((counts.sum()**0.5)/len(data))
    return 1 - sum(counts)**0.5/len(data)

def validate_resource_id(resid):
    """Returns True if resid is valid, False otherwise. Note that validity
       does not imply the resource currently exists in an archive.
       (use .get for that.)"""
    return 0x0100 <= resid <= 0xFFFF
def resid(subindex, n=0):
    """Given a major (master) index page subindex and minor (page) index n,
       return the resource ID."""
    if isinstance(subindex,Resource): subindex,n=subindex.subindex,subindex.n
    assert n < 0x100 and subindex < 0x100
    return ((subindex+1)<<8) | n
def master_index(resid):
    """Return the major (master) index page associated with a resource ID."""
    return ((resid & 0xFF00) >> 8) - 1
def index(resid):
    """Return the minor (page) index associated with a resource ID."""
    return resid & 0xFF
def indices(resid):
    """Return the major and minor indicies associatied with a resource ID.
       If resid is infact already a tuple, return it. """
    return resid if isinstance(resid,tuple) else (master_index(
        resid), index(resid))

class ResourceFile(util.BinaryHandler):
    """Implements a simple file-like interface for resources, with the
       binary utility functions of BinaryHandler."""
    def __init__(self, resource):
        self.resource = resource
        self.position = 0
        self.name = repr(self)
    def __repr__(self):
        return '<ResourceFile from %s. P=%d>'%(self.resource, self.position)
    def tell(self):
        return self.position
    def seek(self, offset, whence=0):
        if whence == 2: # Who uses this?
            self.position = len(self.resource.data) - offset
        elif whence == 1:
            self.position += offset
        elif whence == 0:
            self.position = offset
        else:
            assert False, "Illegal seek whence: %d"%whence
        if self.position >= len(self.resource.data):
            raise IndexError, "Bad seek to 0x%08X, size 0x%08X"%(
                offset, len(self.resource.data))
    def eof(self):
        return self.position >= len(self.resource.data)
    def truncate(self, size=None):
        if size is None: size = self.position
        self.resource.data = self.resource.data[:size]
        self.resource.dirty = True
    def readb(self, length=None):
        if length is None:
            rv = self.resource.data[self.position:]
            self.position = len(self.resource.data)
        else:
            rv = self.resource.data[self.position:self.position+length]
            self.position += length
        return rv
    def read(self, length=None):
        return str(self.readb(length))
    def write(self, string):
        self.resource.data[self.position:self.position+len(string)] = string
        self.resource.dirty = True
        self.position += len(string)
    
        

class Resource(object):
    """A an object that represents a an individual resource in 
       an archive. """
    def __init__(self, offset, size, subindex=0, n=0, archive=None):
        self.offset = offset # 0 if never been on disk
        self.data = bytearray(size) # Data in memory
        self.resid = resid(subindex,n)
	self.loaded = not offset
        self.dirty = True
        self.encrypted = None if not self.loaded else False
        self.canon_encryption = self.encrypted
        # Multiple archives may be open.
        self.archive = archive
        # Only needed for debugging and decryption.
        self.subindex = subindex
        self.n = n
    def preview(self): 
        """Generate a human-readable preview."""
        d = filter(
            (lambda x: chr(x) in string.printable and (
                chr(x) not in string.whitespace)), self.get_data())
        return '"%s"'%''.join(map(chr,d[:20]))
        
        
    def human_readable_size(self):
        lend = len(self.data)
        if lend < 9999: return "%d B"%lend
        elif lend < 1e6: return "%.2f kB"%(lend/1000.0)
        else: return "%.2f MB"%(lend/1e6)
    def as_file(self):
        """Return a file-like object representation of the resource.
           If you write to the object, it will indeed change the contents
           of the resource, but in all cases you need to explicitly write
           out the Archive to save changes to disk. It defines read, write,
           seek, tell, truncate, and the various binary helper methods found in 
           delv.util.BinaryHandler."""
        return ResourceFile(self)
     
    def __getitem__(self, n):
        """Return the nth byte of the resource."""
        if not self.loaded: self.load()
        return self.data[n]
    def __setitem__(self, n, v):
        """Set the nth byte of the resource to v."""
        if not self.loaded: self.load()
        self.dirty = True
        self.data[n] = v
    def set_data(self, data):
        """Replace the data of this resource."""
        self.dirty = True
        self.loaded = True
        self.data = bytearray(data)
    def get_data(self):
        """Return the data of this resource as a mutable bytearray.
           If you alter the bytearray, you must manually set .dirty to True."""
        if not self.loaded: self.load()
        return self.data
    def __repr__(self):
        return "<Resource %04X>"%resid(self.subindex,self.n)
    def __str__(self):
        """Return a string of the Resource's contents."""
        if not self.loaded: self.load()
        return str(self.data)
    def __len__(self):
        "Returns the size of the resource data."
        return len(self.data)
    def hint_encryption(self, encrypted, canon_encryption=None):
        self.encrypted = encrypted
        self.canon_encryption = canon_encryption if (
            canon_encryption is not None) else self.encrypted
    def is_encrypted(self):
        """Returns True if this resource is encrypted _in the medium 
           it was loaded from_."""
        if self.encrypted is None: self.load()
        return self.encrypted
    def canonically_encrypted(self):
        """Returns True if this resource is encrypted in the single-file
           archive. If this Resource came from a Delver Archive loaded 
           from a file (or created de novo), this function returns the
           same value as is_encrypted(). However, if we are dealing with
           an archive loaded from a directory that was dumped from an Archive,
           then this returns the original encryption state from that archive.
           (Recall that all resources are saved unencrypted when saving to
           a directory.)"""
        hint = self.archive.canon_encryption_of(self.subindex)
        if hint is not None:
            self.canon_encryption = hint
            return hint
        if self.canon_encryption is None: self.load()
        return self.canon_encryption
    # Private methods
    def write(self, dest):
        new_offset = dest.tell()
        if not self.loaded: 
            self.load()
            #dest.seek(new_offset)
        self.offset = new_offset
        if self.canon_encryption is True:
            dest.write(encrypt(self.data, resid(self.subindex,self.n)))
        elif self.canon_encryption is False:
            dest.write(self.data)
        else:
            assert False,"Writing %s - undefined encryption status"%repr(self)
    def get_metadata(self):
        return self.canon_encryption
    def write_index(self,dest):
        dest.write_offlen(self.offset, len(self.data))
    def save_to_file(self, dest):
        """Saves a resource to an individual file. Never encrypted."""
        if not self.loaded: self.load()
        dest.seek(0)
        dest.write(self.data)
        self.dirty = False  
    def load_from_file(self, src):
        self.data = bytearray(src.read())
        self.loaded = True
        self.encrypted = False
        self.canon_encryption = self.archive.canon_encryption_of(self.subindex)
    def load(self):
        fpos = self.archive.arcfile.tell()
        self.archive.arcfile.seek(self.offset)
        self.data = bytearray(self.archive.arcfile.read(len(self.data)))
        self.loaded=True
        self.archive.arcfile.seek(fpos)
        if self.encrypted is None:
            self.encrypted = self.archive.canon_encryption_of(self.subindex)
            if self.canon_encryption is None: 
                self.canon_encryption = self.encrypted
        if self.encrypted is None:
            self.decrypt_if_required()
        elif self.encrypted is True:
            self.data = decrypt(self.data, resid(self.subindex,self.n))
        
        
    def decrypt_if_required(self):
        print "Decrypt if required", repr(self), self.archive
        presumptive = decrypt(self.data, resid(self.subindex, self.n))
        if entropy(self.data) > entropy(presumptive):
            self.encrypted = True
            self.canon_encryption = True
            self.data = presumptive
        else:
            self.encrypted = False
            self.canon_encryption = False
        

class Archive(object):
    """Class for representing Delver Archives. The implementation is 
       eager; the entire file is loaded into memory when it is opened."""
    known_encrypted = []
    known_clear = []
    def __init__(self, src=None, archive_type='scenario', gui_treestore=None):
        """If src is None, then the constructor creates a new empty archive.
           If src is a file-like object, it will read in an archive from 
           that file. If src is a string, it will attempt to open it as
           a file read-only."""
        self.gui_treestore=gui_treestore
        self.player_name = ''
        self.encryption_knowledge = {}
        for si in self.known_encrypted: self.encryption_knowledge[si] = True
        for si in self.known_clear: self.encryption_knowledge[si] = False

        if type(src) is str:
            if os.path.isfile(src):
                self.from_file(open(src, 'rb'))
            elif os.path.isdir(src):
                self.from_directory(src)
            else:
                assert False, "'%s' isn't a directory or file."%src
        elif src is None:
            self.arcfile = None
            self.source_string = 'Created de novo by delv'
            self.create_header()
            self.create_index()
        else:
            self.from_file(src)
    def from_directory(self, path):
        """Load a delver archive from the provided path."""
        metadata = json.load(open(os.path.join(path, "metadata.json"),'r'))
        self.source_string = metadata['source']+' (From directory %s)'%path
        encrypt = metadata['should_encrypt']
        self.create_header()
        self.create_index()
        self.scenario_title = metadata['scenario_title']
        self.player_name = metadata['player_name']
        for rid,canon_e in encrypt.items():
            rid = int(rid,16)
            subindex,ri = indices(rid)
            nres = Resource(0,0,subindex,ri, self)
            dfile = open(os.path.join(path, '%04X.data'%rid), 'r')
            nres.load_from_file(dfile)
            nres.hint_encryption(False, canon_e)
            self[rid] = nres
            
        self.source_string = 'Packed from %s'%path
        if self.gui_treestore: self.add_gui_tree()
    def from_file(self, src): 
        """Load a delver archive from a file-like object."""
        self.arcfile = util.BinaryHandler(src)
        self.load_header()
        self.load_index()
        self.source_string = "Loaded from file object %s"%src
    def from_string(self, src):
        """Load a delver archive from an indexable object."""
        self.from_file(cStringIO.StringIO(src))
        self.source_string = "Loaded from string"
    def from_path(self, path):
        """Load a Delver archive from a file system path."""
        self.source_string = "Loaded from path %s"%path
        if os.path.isdir(path):
            self.from_directory(path)
        else:
            self.from_file(open(path, 'rb'))
    def source(self):
        '''Return a string explaining where this archive came from.'''
        return self.source_string
    def to_directory(self, path):
        """Dump the contents of the archive, unencrypted, to path."""
        assert os.path.isdir(path)
        metadata = {'source': self.source(), 
            'creator': 'delv (www.ferazelhosting.net/wiki/delv)',
            'scenario_title': self.scenario_title,
            'player_name': self.player_name}
        encrypt = {}
        for resource in self.resources():
            if not resource: continue # don't preserve empties
            outf = open(os.path.join(path, "%04X.data"%resid(resource)), 'w')
            resource.save_to_file(outf)
            encrypt["%04X"%resid(resource)] = resource.get_metadata()
            outf.close()
        metadata['should_encrypt'] = encrypt
        metafile = open(os.path.join(path, "metadata.json"), 'w')
        json.dump(metadata, metafile)
        metafile.close
    def to_file(self, dest):
        """Write a Delver archive to the destination file-like
           object (must be open for writing, obviously)"""
        dest = util.BinaryHandler(dest)
        self.save_header(dest)
        # Skip to just past the spot where we'll put the master index later
        dest.write(bytearray(0x800))
        # Write all the resources. 
        for n,subindex in enumerate(self.all_subindices):
            if not subindex:
                self.master_index[n] = 0,0
                continue
            at_least_one = False
            for m,res in enumerate(subindex):
                if res: 
                    res.write(dest)
                    at_least_one = True
            if not at_least_one: # Prune empty subindices, if any arise
                self.master_index[n] = 0,0
                self.all_subindices[n] = []
        # Write all our resource indices, now that they know where they are.
        for n,subindex in enumerate(self.all_subindices):
            if not subindex: continue

            # Can make it only as long as it needs to be...?
            self.master_index[n] = dest.tell(), 2048 
            for res in subindex:
                if res:
                    res.write_index(dest)
                else:
                    dest.write_offlen(0,0)
        # Now write the master index:
        dest.seek(self.master_index_offset + 8)
        for offset,length in self.master_index:
            dest.write_offlen(offset,length)
                
    def canon_encryption_of(self, subindex):
        return self.encryption_knowledge.get(subindex, None)

    def to_string(self):
        """Produces one (possibly very large) string with the 
           archive in it. Mainly here for front-end web stuff."""
        stio = StringIO.StringIO()
        self.to_file(stio)
        return stio.getvalue()

    def to_path(self, path):
        """Save the archive to a location on disk."""
        if os.path.isdir(path):
            self.to_directory(path)
        else:
            self.to_file(open(path, 'wb'))

    def get(self, idx, create_new=False):
        """Get a resource by its two-byte-integer composite resource ID,
           or by a tuple of integers (subindex, ID).
           A Resource object is returned if the resource exists, otherwise
           None is returned. (This can be efficiently used to check if a
           Resource exists.)"""
        subindex,n = indices(idx)
        if not self.all_subindices[subindex]:
            if create_new:
                self.all_subindices[subindex] = [None]*256
            else:
                return None
        r = self.all_subindices[subindex][n]
        if r is None and create_new:
            r = Resource(0,0,subindex,n,self)
            self.all_subindices[subindex][n] = r
        return r
                

    def __getitem__(self, idx): 
        """Retrieve the data of a resource using the index[] operator. 
           Its sematics differ slightly from .get in that it will raise 
           IndexError if the resource does not exist, whereas .get will not 
           produce exceptions. idx can be the two-byte resource ID or it can be
           a tuple with the subindex and individual resource ID. Another
           difference is that the content of the resource, and not the
           corresponding Resource object, is returned."""
        res = self.get(idx)
        if res is None:
            raise IndexError("Resource %s is not in use"%idx)
        return res.get_data()
        
    def __setitem__(self, idx, value):
        """idx sematics are the same as __getitem__. If value is not 
           already a Resource object, it will be automatically wrapped.
           New resources are created if needed."""
        subindex,n = indices(idx)
        if isinstance(value, Resource):
            if not self.all_subindices[subindex]:
                self.all_subindices[subindex] = [None]*256
            if not self.master_index[subindex]:
                self.master_index[subindex] = (-1,-1)
            self.all_subindices[subindex][n] = value
            if not value.loaded: value.load()
            value.archive = self
        else:
            res = self.get(idx, True)
            res.set_data(value)
    def resource_ids(self, subindex=None):
        """If subindex is not None, returns a list of valid resource IDs
           (not resource objects) for that subindex, possibly empty if
           the subindex is empty or does not exist. If subindex is None,
           it returns a list of all resource IDs that are valid for this
           archive in toto."""
        if subindex is not None:
            sx = self.all_subindices[subindex]
            return [resid(subindex,n) for n,r in enumerate(sx) if r]
        else:
            return reduce(operator.add, 
                [self.resource_ids(si) for si in self.subindices()])
    def subindices(self):
        """Return a list of valid subindices for this archive."""
        return [n for n in xrange(len(self.master_index)
            ) if self.master_index[n]]


    def resources(self, subindex=None):
        """Returns a list of all the Resource objects in the subindex
           provided. (Possibly an empty list.) If no subindex is provided,
           returns a list of all the resources in the archvie."""
        if subindex is not None:
            sx = self.all_subindices[subindex]
            return [r for r in sx if r]
        else:
            return reduce(operator.add, 
                [self.resources(si) for si in self.subindices()])
    def __iter__(self):
        """This iterator is over all extant resources in the archive. That
           is, if d is an Archive, the following are equivalent:
           for resource in d: ...
           for resource in d.resources(): ...
           Perhaps a future lazy version could be made, in which case the
           iterator would be more efficient.
        """
        return iter(self.resources())
    # --------------------------- Private methods ------------------------
    def create_header(self, scenario="Cythera: Fate of Alaric"):
        self.scenario_title = scenario
        # What these are for, we don't know.
        self.unknown_40 = 0x13
        self.unknown_42 = 2
        self.unknown_48 = 2
        self.master_index_offset = 0x80
        self.master_index_length = 2048
    def create_index(self, size=256):
        self.master_index = [None]*size
        self.all_subindices = []
        for _ in xrange(size): self.all_subindices.append([])
        self.master_index_length = 8*size
    def load_header(self):
        self.scenario_title = self.arcfile.read_pstring(0)
        self.player_name = self.arcfile.read_pstring(0x20)
        self.unknown_40 = self.arcfile.read_uint8(0x40)
        self.unknown_42 = self.arcfile.read_uint8(0x42)
        self.unknown_48 = self.arcfile.read_uint8(0x48)
        self.master_index_offset = self.arcfile.read_uint32(0x80)
        self.master_index_length = self.arcfile.read_uint32(0x84)
    def save_header(self,dest):
        dest.write_pstring(self.scenario_title, 0)
        dest.write_pstring(self.player_name, 0x20)
        dest.write_uint8(self.unknown_40, 0x40)
        dest.write_uint8(self.unknown_42, 0x42)
        dest.write_uint8(self.unknown_48, 0x48)
        dest.write_offlen(self.master_index_offset, self.master_index_length,
            self.master_index_offset)
    def load_index(self):
        self.arcfile.seek(self.master_index_offset+8)
        self.master_index = [None]*(self.master_index_length//8 - 1)
        for n in xrange(len(self.master_index)):
             self.master_index[n] = self.arcfile.read_offlen()
        self.all_subindices = []
        for subn,(offset, length) in enumerate(self.master_index):
            
            if not offset: 
                self.all_subindices.append([])
                continue
            subindex = []
            self.arcfile.seek(offset)
            size = length / 8
            rescount = 0
            for n in xrange(size):
                res_offset, res_length = self.arcfile.read_offlen()
                if res_offset:
                    subindex.append(Resource(res_offset, res_length, 
                       subn, n, self))
                    rescount += 1
                else:
                    subindex.append(None)
            self.all_subindices.append(subindex)
        if self.gui_treestore: self.add_gui_tree()
    def add_gui_tree(self):
        self.gui_tree_rows = {}
        for subn,subindex in enumerate(self.all_subindices):
             rescount = len([r for r in subindex if r])
             if not rescount: continue
             t=self.gui_treestore.append(None,["%3d [%02Xxx]"%(subn,subn+1),
                 "%3d item%s"%(rescount, '' if rescount == 1 else 's'),
                   _SCEN_HINTS.get(subn, "Unknown"), subn,-1])
             self.gui_tree_rows[subn] = t
             for r in subindex:
                 if not r: continue
                 self.gui_treestore.append(t, [
                     "%04X"%resid(r.subindex,r.n),
                     r.human_readable_size(),
                     _RES_HINTS.get(resid(r.subindex,r.n),""),
                     r.subindex,r.n
                     ])

    
        
class Player(Archive):
    """Class for manipulating Delver Player Files, i.e. saved games."""
    

class Scenario(Archive):
    """Class for manipulating Delver Scenario files."""
    known_encrypted = [1,2,4,7,8,9,10,11,12,13,14,15,16,
                       19,20,23,24,25,26,27,29,47]
    known_clear = [0,3,127,128,131,135,137,141,142,143,
                   144,187,239,254]

class Patch(Scenario):
    """Class for manipulating Delver patchfiles."""
    # public:
    patch_info = "This patch created by delv.archive.Patch"
    def patch_info(self, infostring):
        "Set the description of the patch."
        self[0xFFFF] = 'MAGPY'+infostring
        self.patch_info = infostring
    def get_patch_info(self):
        """Return the patch info string. mag.py and Magpie formats supported."""
        patchres = self.get(0xFFFF)
        if not patchres: return ''
        data = patchres.as_file()
        if data.read(5) == 'MAGPY': # mag.py format
            self.patch_info = data.read()
        else: # Magpie format
            self.patch_info = data.read_pstring(0x138)
        return self.patch_info
        
        
    def compatible(self, other_patch, exclude = [0xFFFF,0xBC00,0xBC35]):
        """Returns False if two patches are not compatible. Returns True
           if they might be compatible. ;) """
        exclude = set(exclude)
        set_res = set(self.resource_ids())
        patch_res = set(other_patch.resource_ids())
        return not (set_res-exclude).intersection(patch_res-exclude)
    def patch(self, target):
        "Apply this patch to the target."
        for resource in self:
            target[resid(resource)] = resource
    def diff(self, base, modified,exclude=[0xFFFF]):
        """Add the differences between Archives base and modified to this
           patch archive, such that if .apply(base) is subsequently used,
           base will come to contain the same data as modified."""
        exclude = set(exclude)
        for resid in modified.resource_ids():
            if resid in exclude: continue
            eq = base.get(resid)
            if (not eq) or (eq.get_data() != modified.get(resid).get_data()):
                nd = modified.get(resid)
                if not nd.loaded: nd.load() 
                self[resid] = nd
         
            



#!/usr/bin/env python
# Copyright 2015 Bryce Schroeder, www.bryce.pw, bryce.schroeder@gmail.com
# Wiki: http://www.ferazelhosting.net/wiki/delv
# 
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# "Cythera" and "Delver" are trademarks of either Glenn Andreas or 
# Ambrosia Software, Inc. 

from distutils.core import setup
import setuptools

setup(name='delv',
      version='0.1.13',
      description='Module for making tools interoperable with Delver',
      author='Bryce Schroeder',
      author_email='bryce.schroeder@gmail.com',
      license='GPL-3',
      url='http://www.ferazelhosting.net/wiki/delv',
	classifiers=["Development Status :: 2 - Pre-Alpha",
                     "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
                     "Topic :: Games/Entertainment :: Role-Playing"],
      packages=['delv'])
